package anywheresoftware.b4a.agraham.svg;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;


import com.larvalabs.svgandroid.SVG;
import com.larvalabs.svgandroid.SVGParser;

import android.graphics.Picture;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.drawable.Drawable;
import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.objects.drawable.CanvasWrapper;
import anywheresoftware.b4a.objects.drawable.CanvasWrapper.RectWrapper;
import anywheresoftware.b4a.objects.streams.File;

/**
 *This is a compact and straightforward object for parsing simple SVG files and rendering them
 *in an Android Canvas or as a PictureDrawable that may be assigned to the Background of a View.
 */

@ShortName("SVG")
@Author("Andrew Graham")
@Version(1.5f)
public class SVGWrapper
{
	// 1.0 	initial release
	// 1.1 	external source code for svg-android-2 now compiled into library
	// 1.2 	fixed typo in SVGHandler.doStroke that used "stroke-width" instead of "stroke-linecap"
	// 1.3	added SVGWrapper.CreateFromStream and SVGWrapper.WriteToStream 
	//		made public constructor SVG(Picture, RectF) 
	//		added Q and T quadratic bezier to SVGParser.doPath
	//		made modifications to properly support paint inheritance
	//		removed style attribute
	// 1.4	speeded up attribute searches with a HashMap
	//		re-implemented "style"
	// 1.5	set missing attribute values like x, y, width etc to 0 avoiding puzzling exceptions
	
	private final static double version = 1.5;

	/**
	 *Returns the version number of the library.
	 */
	public double getVersion()
	{
		return version;
	}
	
	/**
	 *This library supports a subset of the SVG Basic 1.1 specification.
	 *Typically, you can just load your vector artwork in a suitable program and then save it
	 *as a SVG Basic 1.1 file and it will work fine.
	 *
	 *By using vector art, the pain of supporting various screen sizes and densities in Android can be reduced. 
	 * 
	 *These are the features of SVG Basic not supported: 
	 *  Raster images (bitmaps).
	 *  Symbols, conditional processing. 
	 *  Patterns.
	 *  Masks, filters and views.
	 *  Interactivity, linking, scripting and animation. 
	 *
	 *Bounds and Limits
	 *It is useful when working with an SVG to know the rectangular bounds of the content.
	 *The document (artboard) bounds themselves are typically not useful for this, as few
	 *artists will size the content to those bounds. Instead, we offer two approaches:
	 *
	 *Bounds
	 *If there is a layer in the artwork called "bounds", it will not be drawn.
	 *Rather, the first rectangle found in that layer will be interpreted as the bounds of the artwork.
	 *This makes it easy for an artist to specify the right bounds right in the artwork without the
	 *developer having to do additional work.
	 * 
	 *Limits
	 *While the parser is reading the SVG, it estimates the bounds of the artwork from what is read.
	 *This may not always be correct, but can often be relied upon for simple artwork. 
	 * 
	 *  
	 *Version 1.0 of this library was a very thin wrapper of the svg-android library which it required
	 *as an external component to be referenced as a Lib by Basic4android.
	 *  <link>svg-android|http://code.google.com/p/svg-android/</link>
	 *  
	 *This version of the library was originally a very thin wrapper of the svg-android-2 library and
	 * the source code for that is compiled into this library so no external component is needed.
	 *This fork is used as the original author seems to have not reponded to bug fix requests so another
	 *programmer has created this fork and applied some fixes and improvements.
	 * Support transforms in g tags
	 * Support transform lists (e.g. transform="rotate(..) scale(..) ...")
	 * Support for arcs in paths 
	 * Support implied commands in paths 
	 * Support for named colors
	 * Support for text (limited, but including font-size, type-face, and text alignment)
	 * Support for stroke-dasharray for dotted/dashed lines
	 * <link>svg-android-2|http://code.google.com/p/svg-android-2/</link>
	 *  
	 *The svg-android library is released for use under the Apache License Version 2.0.
	 *  <link>Apache 2.0|http://www.apache.org/licenses/LICENSE-2.0</link>
	 * 
	 *I have further extensively modified the source code of the original fork to increase performance,
	 *rationalise the structure of the somewhat patchwork original, wring out some bugs (there's a mixed metaphor!)
	 *and incorporate some obvious omissions.
	 *
	 *Unfotunately the svg-android library is not always very informative about SVG files that it cannot
	 *parse. Sometimes it will just throw a NullPointerException without any further information as to
	 *where in the SVG file the problem lies. Other times it may throw an exception whose message
	 *includes the line and column where the error is located in the SVG data file. 
	 */
	public void LIBRARY_DOC()
	{
	}	
	
	private SVG svg;
	
	/**
	 *Initialize the SVG object with a picture obtained from a file name.
	 */
	public void Initialize(String dir, String filename) throws IOException
	{
		File.InputStreamWrapper in = File.OpenInput(dir, filename);
		Initialize2((InputStream)in.getObject());
	    in.Close();
	}
	
	/**
	 *Initialize the SVG object with a picture read from an InputStream.
	 */
	public void Initialize2(InputStream inputstream)
    {
		svg = SVGParser.getSVGFromInputStream(inputstream);
    }
	
	/**
	 *Initialize the SVG object with a picture obtained read from an XML string.
	 */
	public void Initialize3(String svgstring)
    {
		svg = SVGParser.getSVGFromString(svgstring);		
    }
	
	/**
	 *Returns whether a picture exists within the SVG object or not
	 */
	public boolean IsInitialized()
	{
		return svg != null;
	}
	
	/**
	 *Returns an Android PictureDrawable created from the the SVG picture data.
	 *This may be assigned to the Background property of a View for display.
	 *However DrawPicture may be more useful in practice. 
	 */
	public Drawable CreatePictureDrawable()
	{
		return svg.createPictureDrawable();
	}
	
	/**
	 *Draw this picture on the specified canvas to fit into the dest rectangle. 
	 */
	public void DrawPicture(CanvasWrapper canvas, Rect destrect)
	{
		
		canvas.canvas.drawPicture(svg.getPicture(), destrect);
	}
	
	/**
	 *Return the height of the rendered SVG picture.
	 */
	public int getHeight()
	{
		return svg.getPicture().getHeight();
	}

	/**
	 *Returns an Android Picure parsed from the SVG picture data.
	 *This is provided in case of need and is not directly usable in Basic4android. 
	 */
	public Picture getPicture()
	{
		return svg.getPicture();
	}
	
	/**
	 *Return the width of the rendered SVG picture.
	 */
	public int getWidth()
	{
		return svg.getPicture().getWidth();
	}
	
	/**
	 *Gets the bounding rectangle for the SVG that was computed upon parsing.
	 *It may not be entirely accurate for certain curves or transformations
	 *but is often better than nothing. 
	 *Returns a Rect object.
	 */
	public RectWrapper getLimits()
	{
		return recthelper(svg.getLimits());
	}

	/**
	 *Gets the bounding rectangle for the SVG picture data, if one was specified.
	 *Returns a Rect object whose parameters are all 0 if no bounding rectangle was specified.
	 */
	public RectWrapper getBounds()
	{
		return recthelper(svg.getBounds());
	}
	
	private RectWrapper recthelper(RectF rectf)
	{
		RectWrapper rect = new RectWrapper();
		if (rectf != null)
			rect.Initialize((int)rectf.left, (int)rectf.top, (int)rectf.right, (int)rectf.bottom);
		else
			rect.Initialize(0,0,0,0);
		return rect;
	}
	
	/**
	 *Create a new picture (already recorded) from the data in the stream.
	 *This data was generated by a previous call to writeToStream().
	 */
	public void CreateFromStream(InputStream istream)
	{
		Picture p = Picture.createFromStream(istream);
		svg = new SVG(p, new RectF(0, 0, 0, 0));
	}
	
	/**
	 *Write the picture contents to a stream.
	 *The data can be used to recreate the picture by calling createFromStream.
	 *Bounds and Limits are not set
	 */
	public void WriteToStream(OutputStream ostream)
	{
		svg.getPicture().writeToStream(ostream);
	}
	
 }

	
